"""
Tests for NLP
"""
import unittest
from pathlib import Path
import csv

from pytest import approx

from amilib.file_lib import FileLib
from amilib.util import Util
from test.resources import Resources
from test.test_all import AmiAnyTest

from amilib.ami_nlp import AmiNLP, AmiRake

import nltk
nltk.download('stopwords')

logger = Util.get_logger(__name__)

class NLPTest(AmiAnyTest):
    """
    Tests for ami_nlp
    """
    # import nltk, string

    @unittest.skip("needs installation")
    def test_compute_text_similarity_STAT(self):
        """
        ami_nlp,cosine_sim compares strings
        :return:
        """
        ami_nlp = AmiNLP()
        assert ami_nlp.cosine_sim('a little bird', 'a little bird') == approx(1.0)
        assert ami_nlp.cosine_sim('a little bird', 'a little bird chirps') == approx(0.7093, abs=0.001)
        assert ami_nlp.cosine_sim('a little bird', 'a big dog barks') == approx(0, abs=0.001)

    @unittest.skip("crashes on PMR machine with 'illegal instruction' - probably incompatible hardware")
    def test_keyBERT(self):
        """
        from keyBERT docs
        """
        logger.info("start")
        from keybert import KeyBERT

        doc = """
                 Supervised learning is the machine learning task of learning a function that
                 maps an input to an output based on example input-output pairs. It infers a
                 function from labeled training data consisting of a set of training examples.
                 In supervised learning, each example is a pair consisting of an input object
                 (typically a vector) and a desired output value (also called the supervisory signal).
                 A supervised learning algorithm analyzes the training data and produces an inferred function,
                 which can be used for mapping new examples. An optimal scenario will allow for the
                 algorithm to correctly determine the class labels for unseen instances. This requires
                 the learning algorithm to generalize from the training data to unseen situations in a
                 'reasonable' way (see inductive bias).
              """
        kw_model = KeyBERT()
        keywords = kw_model.extract_keywords(doc)
        assert keywords is not None

    def test_rake_nltk(self):
        """
        example from repo/pypi
        """
        ami_rake = AmiRake()

        doc = """
                 Supervised learning is the machine learning task of learning a function that
                 maps an input to an output based on example input-output pairs. It infers a
                 function from labeled training data consisting of a set of training examples.
                 In supervised learning, each example is a pair consisting of an input object
                 (typically a vector) and a desired output value (also called the supervisory signal).
                 A supervised learning algorithm analyzes the training data and produces an inferred function,
                 which can be used for mapping new examples. An optimal scenario will allow for the
                 algorithm to correctly determine the class labels for unseen instances. This requires
                 the learning algorithm to generalize from the training data to unseen situations in a
                 'reasonable' way (see inductive bias).
              """
        ami_rake.read_text(doc)
        ranked_phrases = ami_rake.get_ranked_phrases()
        assert ranked_phrases == [
    'see inductive bias ).',
 'labeled training data consisting',
 'supervised learning algorithm analyzes',
 'supervisory signal ).',
 'mapping new examples',
 'machine learning task',
 'desired output value',
 'training data',
 'training data',
 'training examples',
 'supervised learning',
 'supervised learning',
 'pair consisting',
 'learning algorithm',
 'output pairs',
 'output based',
 'unseen situations',
 'unseen instances',
 'optimal scenario',
 'correctly determine',
 'class labels',
 'also called',
 'input object',
 'inferred function',
 'example input',
 'learning',
 'algorithm',
 'input',
 'example',
 'function',
 'function',
 'way',
 'vector',
 'used',
 'typically',
 'set',
 'requires',
 'reasonable',
 'produces',
 'maps',
 'infers',
 'generalize',
 'allow']

        ranked_phrases_with_scores = ami_rake.get_ranked_phrases_with_scores()
        assert ranked_phrases_with_scores == [
 (15.5, 'see inductive bias ).'),
 (12.166666666666666, 'labeled training data consisting'),
 (11.333333333333334, 'supervised learning algorithm analyzes'),
 (9.5, 'supervisory signal ).'),
 (8.5, 'mapping new examples'),
 (8.333333333333334, 'machine learning task'),
 (8.333333333333334, 'desired output value'),
 (5.166666666666666, 'training data'),
 (5.166666666666666, 'training data'),
 (5.0, 'training examples'),
 (5.0, 'supervised learning'),
 (5.0, 'supervised learning'),
 (5.0, 'pair consisting'),
 (4.666666666666667, 'learning algorithm'),
 (4.333333333333334, 'output pairs'),
 (4.333333333333334, 'output based'),
 (4.0, 'unseen situations'),
 (4.0, 'unseen instances'),
 (4.0, 'optimal scenario'),
 (4.0, 'correctly determine'),
 (4.0, 'class labels'),
 (4.0, 'also called'),
 (3.666666666666667, 'input object'),
 (3.333333333333333, 'inferred function'),
 (3.166666666666667, 'example input'),
 (2.3333333333333335, 'learning'),
 (2.3333333333333335, 'algorithm'),
 (1.6666666666666667, 'input'),
 (1.5, 'example'),
 (1.3333333333333333, 'function'),
 (1.3333333333333333, 'function'),
 (1.0, 'way'),
 (1.0, 'vector'),
 (1.0, 'used'),
 (1.0, 'typically'),
 (1.0, 'set'),
 (1.0, 'requires'),
 (1.0, 'reasonable'),
 (1.0, 'produces'),
 (1.0, 'maps'),
 (1.0, 'infers'),
 (1.0, 'generalize'),
 (1.0, 'allow')]

    def test_rake_wg3_chap03_exec_summ(self):
        """
        IPCC WG3 Chapter03
        """
        text = """
        Chapter 3 assesses the emissions pathways literature in order to identify their key characteristics (both in commonalities and differences) and to understand how societal choices may steer the system into a particular direction (high confidence) . More than 2000 quantitative emissions pathways were submitted to the IPCC’s Sixth Assessment Report AR6 scenarios database, out of which 1202 scenarios included sufficient information for assessing the associated warming consistent with WGI. Five Illustrative Mitigation Pathways (IMPs) were selected, each emphasising a different scenario element as its defining feature: heavy reliance on renewables (IMP-Ren), strong emphasis on energy demand reductions (IMP-LD), extensive use of carbon dioxide removal (CDR) in the energy and the industry sectors to achieve net negative emissions (IMP-Neg), mitigation in the context of broader sustainable development (IMP-SP), and the implications of a less rapid and gradual strengthening of near-term mitigation actions (IMP-GS). {3.2, 3.3}
Pathways consistent with the implementation and extrapolation of countries’ implemented policies until the end of 2020 see greenhouse gas (GHG) emissions reaching 54–61GtCO2-eq yr–1 by 2030 and to 47–67 GtCO2-eq yr–1 by 2050, leading to a median global warming of 2.2°C to 3.5°C by 2100 (medium confidence). These pathways consider policies at the time that they were developed. The Shared Socio-economic Pathways (SSPs) permit a more systematic assessment of future GHG emissions and their uncertainties than was possible in AR5. The main emissions drivers include growth in population, reaching 8.5–9.7 billion by 2050, and an increase in global GDP of 2.7–4.1% per year between 2015 and 2050. Final energy demand in the absence of any new climate policies is projected to grow to around 480–750 EJ yr –1 in 2050 (compared to around 390 EJ in 2015) (medium confidence). The highest emissions scenarios in the literature result in global warming of >5°C by 2100, based on assumptions of rapid economic growth and pervasive climate policy failures ( high conf idence). {3.3}
Many pathways in the literature show how tolimit global warming compared to pre-industrial times to 2°C (>67%) with no overshoot or to limit warming to 1.5°C (>50%) with no or limited overshoot. The likelihood of limiting warming to 1.5°C with no or limited overshoot has dropped in AR6 compared to theSpecial Report on Global Warming of 1.5°C (SR1.5) because global GHG emissions have risen since the time SR1.5 was published, leading to higher near-term emissions (2030) and higher cumulative CO2 emissions until the time of net zero (medium confidence) . Only a small number of published pathways limit global warming to 1.5°C without overshoot over the course of the 21st century. {3.3, Annex III.II.3}
Cost-effective mitigation pathways assuming immediate action1 tolimit warming to 2°C (>67%) are associated with net global GHG emissions of 30–49GtCO2-eq yr–1by 2030 and 14–26GtCO2-eq yr–1by 2050 (medium confidence). This corresponds to reductions, relative to 2019 levels, of 13–45% by 2030 and 52–76% by 2050. Pathways that limit global warming to below 1.5°C with no or limited overshoot require a further acceleration in the pace of the transformation, with net GHG emissions typically around 21–36 GtCO2-eq yr –1 by 2030 and 1–15 GtCO2-eq yr –1 by 2050; thus, reductions of 34–60% by 2030 and 73–98% by 2050 relative to 2019 levels. {3.3}
Pathways following Nationally Determined Contributions (NDCs) announced prior to COP262 until 2030 reach annual emissions of 47–57GtCO2-eq by 2030, thereby making it impossible to limit warming to 1.5°C with no or limited overshoot and strongly increasing the challenge to limit warming to 2°C (>67%) (high confidence). A high overshoot of 1.5°C increases the risks from climate impacts and increases the dependence on large-scale carbon dioxide removal from the atmosphere. A future consistent with NDCs announced prior to COP26 implies higher fossil fuel deployment and lower reliance on low-carbon alternatives until 2030, compared to mitigation pathways with immediate action to limit warming to 2°C (>67%) or lower. To limit warming to 2°C (>67%) after following the NDCs to 2030, the pace of global GHG emission reductions would need to accelerate rapidly from 2030 onward: to an average of 1.4–2.0 GtCO2-eq yr –1 between 2030 and 2050, which is around two-thirds of the global CO2 emission reductions in 2020 due to the COVID-19 pandemic, and around 70% faster than in immediate action pathways that limit warming to 2°C (>67%). Accelerating emission reductions after following an NDC pathway to 2030 would be particularly challenging because of the continued buildup of fossil fuel infrastructure that would be expected to take place between now and 2030. {3.5, 4.2}
Pathways accelerating actions compared to NDCs announced prior to COP26 that reduce annual GHG emissions to48 (38–52) GtCO2-eqby 2030, or 2–9GtCO2-eq below projected emissions from fully implementing NDCs announced prior to COP26, reduce the mitigation challenge forlimiting warming to 2°C (>67%) after 2030 (medium confidence). The accelerated action pathways are characterised by a global, but regionally differentiated, roll out of regulatory and pricing policies. Compared to NDCs, they see less fossil fuels and more low-carbon fuels until 2030, and narrow, but do not close the gap to pathways assuming immediate global action using all available least-cost abatement options. All delayed or accelerated action pathways that limit warming to 2°C (>67%) converge to a global mitigation regime at some point after 2030 by putting a significant value on reducing carbon and other GHG emissions in all sectors and regions. {3.5}
Mitigation pathways limiting warming to 1.5°C (>50%) with no or limited overshoot reach 50% reductions of CO2 in the 2030s, relative to 2019, then reduce emissions further to reach net zero CO2 emissions in the 2050s. Pathwayslimiting warming to 2°C (>67%) reach 50% reductions in the 2040s and net zero CO2by 2070s (medium confidence). {3.3, Cross-Chapter Box 3 in this chapter}
Peak warming in mitigation pathways is determined by the cumulative net CO2 emissions until the time of net zero CO2and the warming contribution of other GHGs and climate forcers at that time (high confidence). Cumulative net CO2 emissions from 2020 to the time of net zero CO2 are 510 (330–710) GtCO2 in pathways that limit warming to 1.5°C (>50%) with no or limited overshoot and 890 (640–1160) GtCO2 in pathways limiting warming to 2°C (>67%). These estimates are consistent with the assessment of remaining carbon budgets by WGI after adjusting for differences in peak warming levels. {3.3, Box 3.4}
Rapid reductions in non-CO2GHGs, particularly methane, would lower the level of peak warming (high confidence). Residual non-CO2 emissions at the time of reaching net zero CO2 range between 5 and 11 GtCO2-eq yr –1 in pathways limiting warming to 2°C (>67%) or lower. Methane (CH4) is reduced by around 19% (4–46%) in 2030 and 45% (29–64%) in 2050, relative to 2019. Methane emission reductions in pathways limiting warming to 1.5°C (>50%) with no or limited overshoot are substantially higher by 2030, 34% (21–57%), but only moderately so by 2050, 51% (35–70%). Methane emissions reductions are thus attainable at relatively lower GHG prices but are at the same time limited in scope in most 1.5°C–2°C pathways. Deeper methane emissions reductions by 2050 could further constrain the peak warming. N2O emissions are reduced too, but similar to CH4, emission reductions saturate for more stringent climate goals. In the mitigation pathways, the emissions of cooling aerosols are reduced due to reduced use of fossil fuels. The overall impact on non-CO2-related warming combines these factors. {3.3}
Net zero GHG emissions imply net negative CO2 emissions at a level compensating residual non-CO2 emissions. Only 30% of the pathways limiting warming to 2°C (>67%) or lower reach net zero GHG emissions in the 21st century (high confidence). In those pathways reaching net zero GHGs, it is achieved around 10 to 40 years later than for net zero CO2 (medium confidence). The reported quantity of residual non-CO2 emissions depends on accounting: the choice of GHG metric. Reaching and sustaining global net zero GHG emissions, measured in terms of GWP-100, results in a gradual decline of temperature ( high confidence). {Cross-Chapter Box 2 in Chapter 2, 3.3, Cross-Chapter Box 3 in this chapter}
Pathways limiting warming to 2°C (>67%) or lower exhibit substantial reductions in emissions from all sectors (high confidence). Projected CO2 emissions reductions between 2019 and 2050 in 1.5°C (>50%) pathways with no or limited overshoot are around 77% (31–96%) for energy demand, 115% (90–167%) for energy supply, and 148% (94–387%) for agriculture, forestry and other land use (AFOLU). In pathways limiting warming to 2°C (>67%), projected CO2 emissions are reduced between 2019 and 2050 by around 49% for energy demand, 97% for energy supply, and 136% for AFOLU (medium confidence). {3.4}
Delaying or sacrificing emissions reductions in one sector or region involves compensating reductions in other sectors or regions if warming is to be limited (high confidence). Mitigation pathways show differences in the timing of decarbonisation and when net zero CO2 emissions are achieved across sectors and regions. At the time of global net zero CO2 emissions, emissions in some sectors and regions are positive while others are negative; the ordering depends on the mitigation options available, the cost of those options, and the policies implemented. In cost-effective mitigation pathways, the energy-supply sector typically reaches net zero CO2 before the economy as a whole, while the demand sectors reach net zero CO2 later, if ever ( high confidence). {3.4}
Pathways limiting warming to 2°C (>67%) or lower involve substantial reductions in fossil fuel consumption and a near elimination of the use of coal without carbon capture and storage (CCS) (high confidence). These pathways show an increase in low-carbon energy, with 88% (69–97%) of primary energy coming from these sources by 2100. {3.4}
Stringent emissions reductions at the level required for 2°C (>67%) or lower are achieved through increased direct electrification of buildings, transport, and industry, resulting in increased electricity generation in all pathways (high confidence). Nearly all electricity in pathways limiting warming to 2°C (>67%) or lower is from low- or no-carbon technologies, with different shares of nuclear, biomass, non-biomass renewables, and fossil CCS across pathways. {3.4}
The measures required tolimit warming to 2°C (>67%) or lower can result in large-scale transformation of the land surface (high confidence). Pathways limiting warming to 2°C (>67%) or lower are projected to reach net zero CO2 emissions in the AFOLU sector between the 2020s and 2070, with an increase of forest cover of about 322 million ha (–67 to 890 million ha) in 2050 in pathways limiting warming to 1.5°C (>50%) with no or limited overshoot. Cropland area to supply biomass for bioenergy (including bioenergy with carbon capture and storage – BECCS) is around 199 (56–482) million ha in 2050 in pathways limiting warming to 1.5°C (>50%) with no or limited overshoot. The use of bioenergy can lead to either increased or reduced emissions, depending on the scale of deployment, conversion technology, fuel displaced, and how/where the biomass is produced ( high confidence). {3.4}
Anthropogenic land CO2 emissions and removals in Integrated Assessment Model (IAM) pathways cannot be directly compared with those reported in national GHG inventories (high confidence). Methodologies enabling a more like-for-like comparison between models’ and countries’ approaches would support more accurate assessment of the collective progress achieved under the Paris Agreement. {3.4, 7.2.2.5}
Pathways that limit warming to 2°C (>67%) or lowerinvolve some amount of CDR to compensate for residual GHG emissions remaining after substantial direct emissions reductions in all sectors and regions (high confidence). CDR deployment in pathways serves multiple purposes: accelerating the pace of emissions reductions, offsetting residual emissions, and creating the option for net negative CO2 emissions in case temperature reductions need to be achieved in the long term ( high confidence). CDR options in the pathways are mostly limited to BECCS, afforestation and direct air carbon capture and storage (DACCS). CDR through some measures in AFOLU can be maintained for decades but not in the very long term because these sinks will ultimately saturate ( high conf idence). {3.4}
Mitigation pathways show reductions in energy demand relative to reference scenarios, through a diverse set of demand-side interventions (high confidence). Bottom-up and non-IAM studies show significant potential for demand-side mitigation. A stronger emphasis on demand-side mitigation implies less dependence on CDR and, consequently, reduced pressure on land and biodiversity. {3.4, 3.7}
Limiting warming requires shifting energy investments away from fossilfuels and towards low-carbon technologies (high confidence). The bulk of investments are needed in medium- and low-income regions. Investment needs in the electricity sector are on average 2.3 trillion USD2015 yr –1 over 2023 to 2052 for pathways that limit warming to 1.5°C (>50%) with no or limited overshoot, and 1.7 trillion USD2015 yr –1 for pathways that limit warming to 2°C (>67%). {3.6.1}
Pathways limiting warming to 2°C (>67%) require more rapid near-term transformations and are associated with higher upfront transition costs, but meanwhile bring long-term gains for the economy as well as earlier benefits in avoided climate change impacts (high confidence). This conclusion is independent of the discount rate applied, though the modelled cost-optimal balance of mitigation action over time does depend on the discount rate. Lower discount rates favour earlier mitigation, reducing reliance on CDR and temperature overshoot. {3.6.1, 3.8}
Mitigation pathwaysthat limit warming to 2°C (>67%) entail losses in global GDP with respect to reference scenarios of between 1.3% and 2.7% in 2050; and in pathways that limit warming to 1.5°C (>50%) with no or limited overshoot, losses are between 2.6% and 4.2%. Yet, these estimates do not account for the economic benefits of avoided climate change impacts (medium confidence). In mitigation pathways that limit warming to 2°C (>67%), marginal abatement costs of carbon are about 90 (60–120) USD2015 tCO2 in 2030 and about 210 (140–340) USD2015 tCO2 in 2050; in pathways that limit warming to 1.5°C (>50%) with no or limited overshoot, they are about 220 (170–290) USD2015 tCO2 in 2030 and about 630 (430–990) USD2015 tCO2 in 2050. 3 {3.6.1}
The global benefits of pathways limiting warming to 2°C (>67%) outweigh global mitigation costs over the 21st century, if aggregated economic impacts of climate change are at the moderate to high end of the assessed range, and a weight consistent with economic theory is given to economic impacts over the longterm. This holds true even without accounting for benefits in other sustainable development dimensions or non-market damages from climate change (medium confidence). The aggregate global economic repercussions of mitigation pathways include the macroeconomic impacts of investments in low-carbon solutions and structural changes away from emitting activities, co-benefits and adverse side effects of mitigation, (avoided) climate change impacts, and (reduced) adaptation costs. Existing quantifications of global aggregate economic impacts show a strong dependence on socio-economic development conditions, as these shape exposure and vulnerability and adaptation opportunities and responses. (Avoided) impacts for poorer households and poorer countries represent a smaller share in aggregate economic quantifications expressed in GDP or monetary terms, whereas their well-being and welfare effects are comparatively larger. When aggregate economic benefits from avoided climate change impacts are accounted for, mitigation is a welfare-enhancing strategy ( high confidence). {3.6.2}
The economic benefits on human health from air quality improvement arising from mitigation action can be of the same order of magnitude as mitigation costs, and potentially even larger (medium confidence). {3.6.3}
Differences between aggregate employment in mitigation pathways compared to reference scenarios are relatively small, although there may be substantial reallocations across sectors, with job creation in some sectors and job losses in others (medium confidence). The net employment effect (and its sign) depends on scenario assumptions, modelling framework, and modelled policy design. Mitigation has implications for employment through multiple channels, each of which impacts geographies, sectors and skill categories differently (medium confidence). {3.6.4}
The economic repercussions of mitigation vary widely across regions and households, depending on policy design and level of international cooperation (high confidence). Delayed global cooperation increases policy costs across regions, especially in those that are relatively carbon intensive at present ( high confidence). Pathways with uniform carbon values show higher mitigation costs in more carbon-intensive regions, in fossil fuel exporting regions and in poorer regions ( high confidence). Aggregate quantifications expressed in GDP or monetary terms undervalue the economic effects on households in poorer countries; the actual effects on welfare and well-being are comparatively larger ( high confidence). Mitigation at the speed and scale required to limit warming to 2°C (>67%) or lower implies deep economic and structural changes, thereby raising multiple types of distributional concerns across regions, income classes and sectors ( high confidence). {3.6.1, 3.6.4}
The timing of mitigation actions and their effectiveness will have significant consequences for broader sustainable development outcomes in the longer term (high confidence). Ambitious mitigation can be considered a precondition for achieving the Sustainable Development Goals (SDGs), especially for vulnerable populations and ecosystems with little capacity to adapt to climate impacts. Dimensions with anticipated co-benefits include health, especially regarding air pollution, clean energy access and water availability. Dimensions with potential trade-offs include food, employment, water stress, and biodiversity, which come under pressure from large-scale CDR deployment, energy affordability/access, and mineral-resource extraction ( high confidence). {3.7}
Many of the potential trade-offs of mitigation measures for other sustainable development outcomes depend on policy design and can thus be compensated or avoided with additional policies and investments or through policies that integrate mitigation with other SDGs (high confidence). Targeted SDG policies and investments, for example in the areas of healthy nutrition, sustainable consumption and production, and international collaboration, can support climate change mitigation policies and resolve or alleviate trade-offs . Trade-offs can be addressed by complementary policies and investments, as well as through the design of cross-sectoral policies integrating mitigation with the Sustainable Development Goals of health, nutrition, sustainable consumption and production, equity and biodiversity. {3.7}
Decent living standards, which encompass many SDG dimensions, are achievable at lower energy use than previously thought (high confidence). Mitigation strategies that focus on lower demands for energy and land-based resources exhibit reduced trade-offs and negative consequences for sustainable development relative to pathways involving either high emissions and climate impacts or those with high consumption and emissions that are ultimately compensated by large quantities of BECCS. {3.7}
Different mitigation pathways are associated with different feasibility challenges, though appropriate enabling conditions can reduce these challenges (high confidence). Feasibility challenges are transient and concentrated in the next two to three decades ( high confidence). They are multidimensional, context-dependent and malleable to policy, technological and societal trends. {3.8}
Mitigation pathways are associated with significant institutional and economic feasibility challenges rather than technological and geophysical feasibility challenges (medium confidence). The rapid pace of technological development and deployment in mitigation pathways is not incompatible with historical records. Institutional capacity is rather a key limiting factor for a successful transition. Emerging economies appear to have the highest feasibility challenges in the short to medium term. {3.8}
Pathways relying on a broad portfolio of mitigation strategies are more robust and resilient (high confidence). Portfolios of technological solutions reduce the feasibility risks associated with the low-carbon transition. {3.8}"""
        ami_rake = AmiRake()
        ami_rake.read_text(text)
        scored_phrases = ami_rake.get_ranked_phrases_with_scores()
        print(scored_phrases[:20])
        assert len(scored_phrases) == 1237
        assert scored_phrases[:20] == [(58.877160104770205, '2 ° c (> 67 %) outweigh global mitigation costs'),
 (57.95613518374528, '2 ° c (> 67 %), marginal abatement costs'),
 (48.49721770370979, '2 ° c (> 67 %), projected co2 emissions'),
 (47.3922902494331, 'net ghg emissions typically around 21 – 36 gtco2'),
 (47.262441490051586, '2 ° c (> 67 %) entail losses'),
 (47.25053672814683, '2 ° c (> 67 %) reach 50'),
 (44.38198388050872, 'net zero ghg emissions imply net negative co2 emissions'),
 (42.59577482338492, '2 ° c (> 67 %) converge'),
 (40.59577482338492, '2 ° c (> 67 %) require'),
 (40.17258297258297, 'around 480 – 750 ej yr – 1'),
 (39.70265143345301, '5 ° c – 2 ° c pathways'),
 (39.61047619047619,
  'effective mitigation pathways assuming immediate action1 tolimit warming'),
 (37.45599715918865, '5 ° c (> 50 %) pathways'),
 (35.74184782608695, 'supply sector typically reaches net zero co2'),
 (35.59577482338492, '2 ° c (> 67 %)'),
 (35.59577482338492, '2 ° c (> 67 %)'),
 (35.59577482338492, '2 ° c (> 67 %)'),
 (35.59577482338492, '2 ° c (> 67 %)'),
 (35.59577482338492, '2 ° c (> 67 %)'),
 (35.59577482338492, '2 ° c (> 67 %)')]


    def test_bloom2_filter(self):
        """
        quick index with adjustable error rate
        """
        from bloom_filter2 import BloomFilter

        # instantiate BloomFilter with custom settings,
        # max_elements is how many elements you expect the filter to hold.
        # error_rate defines accuracy; You can use defaults with
        # `BloomFilter()` without any arguments. Following example
        # is same as defaults:
        bloom = BloomFilter(max_elements=10000, error_rate=0.1)

        # Test whether the bloom-filter has seen a key:
        assert "test-key" not in bloom

        # Mark the key as seen
        bloom.add("test-key")

    @unittest.skip("cannot load/run ImportError: numpy.core.multiarray failed to import")
    def test_pke(self):
        """
        tests PKE - python keyphrase extraction
        https://github.com/boudinfl/pke?tab=readme-ov-file#minimal-example
        """
        import pke

        # initialize keyphrase extraction model, here TopicRank
        extractor = pke.unsupervised.TopicRank()

        # load the content of the document, here document is expected to be a simple
        # test string and preprocessing is carried out using spacy
        extractor.load_document(input='text', language='en')

        # keyphrase candidate selection, in the case of TopicRank: sequences of nouns
        # and adjectives (i.e. `(Noun|Adj)*`)
        extractor.candidate_selection()

        # candidate weighting, in the case of TopicRank: using a random walk algorithm
        extractor.candidate_weighting()

        # N-best selection, keyphrases contains the 10 highest scored candidates as
        # (keyphrase, score) tuples
        keyphrases = extractor.get_n_best(n=10)





"""
downstream parser for pygetpapers
"""
import datetime
import json
from pathlib import Path
import lxml.etree as ET

# from amilib.ami_bib import SAVED, QUERY, STARTDATE, ENDDATE
from amilib.ami_html import HtmlLib, HtmlUtil
from amilib.ami_util import AmiJson, AmiUtil
# from amilib.ami_bib import DOI, AUTHOR_STRING, JOURNAL_INFO_TITLE, PUB_YEAR, ABS_TEXT, SAVED_CONFIG_INI, Pygetpapers
from amilib.util import Util

EUPMC_RESULTS_JSON = "eupmc_results.json"
DATATABLES_HTML = "datatables.html"
SAVED_CONFIG_INI = "saved_config.ini"  # TODO cyclic import

logger = Util.get_logger(__name__)


class AmiCorpus():
    """

    """
    EUPMC_TRANSFORM = {
        "doi": {
            "url": {
                "prefix": "https://www.doi.org/",
            }
        },
        "authorString": {
            "text": {
                "split": ",",
            }
        },
        "abstractText": {
            "text": {
                "truncate": 200,
            }
        },
        "pmcid": {
            "url": {
                "prefix": "https://europepmc.org/betaSearch?query=",
            }

        }

    }

    def __init__(self, indir=None):
        """

        """
        self.indir = indir
        self.eupmc_results = None

    def get_datatables(self):
        """
        create a JQuery datatables from eumpc_reults.json
        """
        self.get_or_create_eupmc_results()
        if self.eupmc_results is None:
            logger.error(f"no {EUPMC_RESULTS_JSON}, so cannot create datatables at present")
            return

    def get_or_create_eupmc_results(self):
        f"""
        gets file {EUPMC_RESULTS_JSON} from directory
        :return: {EUPMC_RESULTS_JSON} from directory or None if not exists
        """
        if self.eumpc_results is None:
            if not Path(self.indir).is_dir():
                logger.error(f"not a directory {self.indir}")
                return None
            self.eupmc_results = Path(self.eupmc_results, EUPMC_RESULTS_JSON)
            if not self.eupmc_results.exists():
                self.eupmc_results = None
        return self.eupmc_results

    @classmethod
    def make_datatables(cls, indir, outdir=None, outfile_h=None):
        """
        creates a JQuery.datatables (HTML) file from an AmiCorpus
        May make this method a member function later
        :param indir: directory with corpus (normally created by pygetpapers)
        :param outdir: output for datatables (if omitted uses indir)
        :param outfile_h: the HTML file ceated (may be changed)
        """
        from amilib.ami_bib import SAVED_CONFIG_INI

        if outdir is None:
            outdir = indir
        if outfile_h is None:
            outfile_h = Path(outdir, DATATABLES_HTML)
        # wanted_keys = [PMCID, DOI, TITLE, AUTHOR_STRING, JOURNAL_INFO_TITLE, PUB_YEAR, ABS_TEXT]
        config_ini = Path(indir, SAVED_CONFIG_INI)
        infile = Path(indir, EUPMC_RESULTS_JSON)
        outdir.mkdir(parents=True, exist_ok=True)
        datatables = True
        AmiCorpus.read_json_create_write_html_table(
            infile, outfile_h, wanted_keys=None, datatables=datatables, table_id=None, config_ini=config_ini)

    @classmethod
    def read_json_create_write_html_table(
            cls, infile, outfile_h, wanted_keys,
            styles=None, datatables=None, table_id=None, config_ini=None):
        """
        read pygetpapers output, select columns and create HTML table
        :param infile: eumpc_results from pygetpapers
        :param outfile_h: HTML table to write
        :param wanted_keys: column headings to select
        :param styles: datatables styles
        :param datatables: datatables
        :param table_id: id links script to table (default "my_table")
        :param config_ini: adds info from saved_config.ini - mainly the query
        :return: html table
        """
        # TODO cyclic imports
        from amilib.ami_bib import PMCID, TITLE
        from amilib.ami_bib import DOI, AUTHOR_STRING, JOURNAL_INFO_TITLE, PUB_YEAR, ABS_TEXT
        if styles is None:
            styles = ["td {border:solid 1px black;}"]
        if table_id is None:
            table_id = "my_table"
        if wanted_keys is None:
            wanted_keys = [PMCID, DOI, TITLE, AUTHOR_STRING, JOURNAL_INFO_TITLE, PUB_YEAR, ABS_TEXT]

        assert infile.exists(), f"infile {infile} must exist"
        with open(infile, "r") as f:
            jsonx = json.load(f)
        # look for all papers
        papers = jsonx.get("papers")
        assert papers is not None, f"cannot find papers"
        # specific keys we want
        dict_by_id = AmiJson.create_json_table(papers, wanted_keys)
        htmlx, table = HtmlLib.create_html_table(
            dict_by_id, transform_dict=AmiCorpus.EUPMC_TRANSFORM,
            styles=styles, datatables=datatables, table_id=table_id
        )
        cls.add_query_as_caption(config_ini, table)

        HtmlUtil.write_html_elem(htmlx, outfile_h, debug=True)

    @classmethod
    def add_query_as_caption(cls, config_ini, table):
        """
        add query string from config.ini to table caption
        :param config_ini: saved_config.ini file
        :param table: to add caption to
        """
        from amilib.ami_bib import SAVED, QUERY, STARTDATE, ENDDATE
        if config_ini is not None:
            config, _ = AmiUtil.get_config_and_section_names(config_ini)
            saved_section = config[SAVED]
            query = saved_section[QUERY]
            startdate = saved_section[STARTDATE]
            if startdate == 'False':
                startdate = "..."
            enddate = saved_section[ENDDATE]
            if enddate == 'False':
                enddate = datetime.date.today()

            caption = ET.SubElement(table, "caption")
            caption.text = f"query: {query}"
            if startdate:
                caption.text += f"; start: {startdate}"
            if enddate:
                caption.text += f"; end: {enddate}"



# test util

import csv
import logging
import re
import shutil
import sys
import unittest
from pathlib import Path

from amilib.ami_util import AmiJson
from amilib.file_lib import FileLib
from amilib.util import EnhancedRegex
from amilib.util import Util, GithubDownloader
from amilib.ami_args import ArgParseBuilder, AmiArgParser, AmiArgParseException
from amilib.xml_lib import Templater

from test.resources import Resources
from test.test_all import AmiAnyTest

# local


logger = Util.get_logger(__name__)

class Util0Test(AmiAnyTest):
    # def __init__(self):
    sys_argv_save = None

    # @classmethod
    # def setUp(cls):
    #     """save args as they will be edited"""
    #     cls.sys_argv_save = sys.argv
    #
    # @classmethod
    # def tearDown(cls):
    #     """restore args"""
    #     sys.argv = cls.sys_argv_save

    @classmethod
    @unittest.skip("not working properly, I think some tests change the args...")
    # TODO fix args - some  tests change the args
    def test_add_argstr(cls):
        # this is a hack as normally there is only one element
        # sys.argv = sys.argv[1:]
        # assert sys.argv[1:] == []
        cmd = "--help foo bar plinge"
        Util.add_sys_argv_str(cmd)
        assert sys.argv[1:] == ["--help", "foo", "bar", "plinge"]

    @classmethod
    @unittest.skip("not working properly")
    # TODO fix args
    def test_add_args(cls):
        # this is a hack as normally there is only one element
        sys.argv = sys.argv[1:]
        # assert sys.argv[1:] == []
        args = ["--help", "foox", "barx", "plingex"]
        Util.add_sys_argv(args)
        assert sys.argv[1:] == ["--help", "foox", "barx", "plingex"]

    @classmethod
    def test_copy_anything(cls):
        src = Resources.TEST_CLIMATE_10_SVG_DIR
        dst = Path(AmiAnyTest.TEMP_DIR, "tempzz")
        if dst.exists():
            shutil.rmtree(dst)
        FileLib.copyanything(src, dst)
        assert Path(dst).exists()

    def test_create_name_value(self):
        """tests parsing of PyAMI flags
        """
        name, value = Util.create_name_value("foo=bar")
        assert name, value == ("foo", "bar")
        name, value = Util.create_name_value("foo")
        assert name, value == ("foo", True)
        try:
            arg = "foo=bar=plugh"
            Util.create_name_value(arg)
            raise ValueError(f"failed to trap {arg}")
        except ValueError as ve:
            assert str(ve == "too many delimiters in {arg}")
        try:
            arg = "foo bar"
            _, v = Util.create_name_value(arg)
            raise ValueError(f"failed to trap {arg}")
        except ValueError as ve:
            assert str(ve) == "arg [foo bar] may not contain whitespace"

        Util.create_name_value("foo/bar")
        assert name, value == "foo/bar"

        Util.create_name_value("foo/bar", delim="/")
        assert name, value == ("foo", "bar")

        assert Util.create_name_value("") is None

        arg = "foo bar"
        try:
            _, v = Util.create_name_value(arg, delim=" ")
            raise ValueError(f"failed to trap {arg}")
        except ValueError as ve:
            assert str(ve) == f"arg [{arg}] may not contain whitespace"

    def test_read_csv(self):
        """
        use Python csv to select column values
        Reads compound_enzyme.csv into CSV DictReader and checks values in first 3 rows
        with column names "NAME" and "TYPE"
        """
        csv_file = Path(Resources.TEST_RESOURCES_DIR, "eoCompound", "compound_enzyme.csv")
        assert csv_file.exists(), f"{csv_file} should exist"
        expected_row_values = [["isopentenyl diphosphate", "COMPOUND"],
                               ["dimethylallyl diphosphate", "COMPOUND"],
                               ["hemiterpene", "COMPOUND"]]
        with open(str(csv_file), newline='') as csvfile:
            reader = csv.DictReader(csvfile)

            for i, row in enumerate(reader):
                if i < 3:
                    assert row['NAME'] == expected_row_values[i][0]
                    assert row['TYPE'] == expected_row_values[i][1]

    def test_select_csv_field_by_type(self):
        """select value in column of csv file by value of defining column
        """
        csv_file = Path(Resources.TEST_RESOURCES_DIR, "eoCompound", "compound_enzyme.csv")
        assert csv_file.exists(), f"{csv_file} should exist"
        selector_column = "TYPE"
        column_with_values = "NAME"
        selector_value = "COMPOUND"

        values = Util.extract_csv_fields(csv_file, column_with_values, selector_column, selector_value)
        assert len(values) == 89
        assert values[:3] == ['isopentenyl diphosphate', 'dimethylallyl diphosphate', 'hemiterpene']

        selector_value = "ENZYME"
        values = Util.extract_csv_fields(csv_file, column_with_values, selector_column, selector_value)
        assert len(values) == 92
        assert values[:3] == ['isomerase', 'GPP synthase', 'FPP synthase']

    def test_create_arg_parse(self):
        arg_parse_file = Path(Resources.TEST_RESOURCES_DIR, "arg_parse.json")
        arg_parse_builder = ArgParseBuilder()
        arg_dict = arg_parse_builder.create_arg_parse(arg_dict_file=arg_parse_file)

    def test_range_list_contains_int(self):
        """does a range or range list contain an int"""
        # single
        rangex = range(1, 3)
        assert not Util.range_list_contains_int(0, rangex)
        assert Util.range_list_contains_int(1, rangex)
        assert not Util.range_list_contains_int(3, rangex)
        # list
        range_list = [range(1, 3), range(5, 9)]
        assert not Util.range_list_contains_int(0, range_list)
        assert Util.range_list_contains_int(1, range_list)
        assert not Util.range_list_contains_int(3, range_list)
        assert not Util.range_list_contains_int(4, range_list)
        assert Util.range_list_contains_int(5, range_list)
        assert not Util.range_list_contains_int(9, range_list)
        assert not Util.range_list_contains_int(10, range_list)
        # None
        range_list = None
        assert not Util.range_list_contains_int(0, range_list)
        range_list = range(1, 3)
        assert not Util.range_list_contains_int(None, range_list)

    def test_get_file_from_url(self):
        url = None
        assert Util.get_file_from_url(url) is None
        url = "https://foo.bar/plugh/bloop.xml"
        assert Util.get_file_from_url(url) == "bloop.xml"

    @unittest.skip("NYI")
    def test_make_id_from_match_and_idgen(self):
        """idgen is of the form <grouo>some text<group>
        where groups correspond to named capture groups in regex
        """
        idgen = "12/CMA.34"
        components = ["", ("decision", "\\d+"), "/", ("type", "CP|CMA|CMP"), "\\.", ("session", "\\d+"), ""]
        enhanced_regex = EnhancedRegex(components=components)
        id = enhanced_regex.make_id(idgen)
        assert id == "12_CMA_34"

    # ABANDONED
    # def test_make_regex_with_capture_groups(self):
    #     """idgen is of the form <grouo>some text<group>
    #     where groups correspond to named capture groups in regex
    #     """
    #     enhanced_regex = EnhancedRegex()
    #     components = ["", ("decision", "\d+"), "/", ("type", "CP|CMA|CMP"), "\.", ("session", "\d+"), ""]
    #     regex = enhanced_regex.make_regex_with_capture_groups(components)
    #     assert regex == '(?P<decision>\\d+)/(?P<type>CP|CMA|CMP)\\.(?P<session>\\d+)'

    def test_make_components_from_regex(self):
        """splits regex with capture groups into its components
        """
        regex = '(?P<decision>\\d+)/(?P<type>CP|CMA|CMP)\\.(?P<session>\\d+)'
        re_parser = EnhancedRegex(regex=regex)
        components = re_parser.make_components_from_regex(regex)
        assert len(components) == 7
        assert components[1] == '(?P<decision>\\d+)'
        assert components[3] == '(?P<type>CP|CMA|CMP)'
        unittest.TestCase().assertListEqual(components,
                                            ['', '(?P<decision>\\d+)', '/', '(?P<type>CP|CMA|CMP)', '\\.',
                                             '(?P<session>\\d+)', ''])

    def test_get_username(self):
        """gets username
        """
        username = Util.get_username()
        if AmiAnyTest.IS_PMR:
            assert username == "pm286"


class GithubDownloaderTest(AmiAnyTest):
    # def __init__(self):
    #     pass

    @unittest.skip("VERY LONG, DOWNLOADS")
    def test_explore_main_page(self):
        owner = "petermr"
        repo = "CEVOpen"
        downloader = GithubDownloader(owner=owner, repo=repo, max_level=1)
        page = None
        downloader.make_get_main_url()
        logger.info(f"main page {downloader.main_url}")
        url = downloader.main_url
        if not url:
            print(f"no page {owner}/{repo}")
            return None

        downloader.load_page(url, level=0)


class AmiArgParserTest(AmiAnyTest):

    def test_ami_arg_parse(self):
        """
        test subclassing of argParse
        """
        ami_argparse = AmiArgParser()
        ami_argparse.add_argument("--flt", type=float, nargs=1, help="a float", default=80)
        ami_argparse.add_argument("--str", type=str, nargs=1, help="a string")

        # this works
        arg_dict = ami_argparse.parse_args(["--flt", "3.2"])
        logger.debug(f"arg_dict1 {arg_dict}")
        # this fails
        try:
            arg_dict = ami_argparse.parse_args(["--flt", "3.2", "str"])
        except AmiArgParseException as se:
            print(f"arg parse error {se} line: {se.__traceback__.tb_lineno}")
        except Exception as e:
            print(f" error {e}")

        arg_dict = ami_argparse.parse_args(["--flt", "99.2"])
        logger.debug(f"arg_dict2 {arg_dict}")


class TemplateTest(AmiAnyTest):

    def test_id_templates(self):
        """Splits files at Decisions"""
        """requires previous test to have been run"""

        template_values = {
            'DecRes': 'Decision',
            'decision': 1,
            'type': "CMA",
            'session': "3",
        }

        template = "{DecRes}_{decision}_{type}_{session}"
        regex = "(?P<DecRes>Decision|Resolution)\\s(?P<decision>\\d+)/(?P<type>CMA|CMP|CP)\\.(?P<session>\\d+)"
        ss = ["Decision 12/CMP.5", "Decision 10/CP.17", "Decision 2/CMA.2", "Decision 4/CCC.9"]
        matched_templates = Templater.get_matched_templates(regex, ss, template)
        assert matched_templates == ['Decision_12_CMP_5', 'Decision_10_CP_17', 'Decision_2_CMA_2', None]

# class UtilTests:
#     def test_dict_read(self):
#         file = "section_templates.json"
#         return Util.read_pydict_from_json(file)

class CommandlineTest:

    def test_command(self):
        command = "--help"

class LoggingTest(AmiAnyTest):
    import logging

    def test_logger(self):
        logger = Util.get_logger(__name__)

        logger.info("Server started listening on port 8080")
        logger.warning(
            "Disk space on drive '/var/log' is running low. Consider freeing up space"
        )
        logger.info("")

        try:
            raise Exception("Failed to connect to database: 'my_db'")
        except Exception as e:
            logger.info(f"=========start Exception============")
            # exc_info=True ensures that a Traceback is included
            FileLib.log_exception(e, logger)
            logger.info(f"=========end Exception============")

        print(f"FINISHED")

class AmiJsonTest(AmiAnyTest):
    """
    tricky JSON stuff (e.g. nested dicts)
    """

    def test_nested_dicts(self):
        """
        access values in nested dict by dot-separated strings
        """
        dikt = {
            "key1" : "value1",
            "key2" :
                {
                    "subkey1": "subvalue21",
                    "subkey2": "subvalue22",

                },
            "key3":
                {
                    "subkey1":
                        {
                            "subsubkey1": "subsubvalue311",
                            "subsubkey2": "subsubvalue312",
                        },
                }

        }
        assert "value1" == AmiJson.read_nested_dicts(dikt, "key1")
        assert "subvalue21" == AmiJson.read_nested_dicts(dikt, "key2.subkey1")
        assert "subsubvalue312" == AmiJson.read_nested_dicts(dikt, "key3.subkey1.subsubkey2")




import re
import unittest
from collections import Counter
from pathlib import Path

import pytest
from geopy.geocoders import Nominatim

from amilib.ami_html import HtmlUtil
from amilib.file_lib import FileLib
from amilib.util import Util
from test.resources import Resources
from test.test_all import AmiAnyTest

subs = {
    "(\\d\\d?:)?\\d\\d?:\\d\\d\\n": "",
    ", , ": " ",
    "\\s*,\\s*": "",
    "OK,?\\s*":" ",
    "Amy\\s+[Dd]ict": "amidict",
    "Amy": "ami",
    "F C C C": "FCCC",
    "So\\s*,?\\s*": " ",
    "\\,\\s*": " ",
    ", ": " ",

}

logger = Util.get_logger(__name__)

def multiple_replace(replacements, text):
    # Create a regular expression from the dictionary keys
    regex = re.compile("(%s)" % "|".join(map(re.escape, replacements.keys())))
    # For each match, look-up corresponding value in dictionary
    return regex.sub(lambda mo: replacements[mo.group()], text)


class MiscTest(AmiAnyTest):

    @unittest.skip("requires connection and output can micro-vary")
    def test_geolocate_GEO(self):
        """
        GEO: locates places by name using Nominatim

        (Test occasionally returns variable decimal places
        TODO needs a fixed-place numeric comparison
        :return:
        """
        geolocator = Nominatim(timeout=10, user_agent="semanticclimate@gmail.com")
        results = []
        for name in [
            "Benares",
#            "Bengaluru",
            "Delhi",
            "Ladakh",
            # "Mumbai",
            "Mysore",
        ]:
            location = geolocator.geocode(name)
            tuple = (name, location[1], location.latitude, location.longitude)
            results.append(tuple)
        assert results == [
            ('Benares', (25.3356491, 83.0076292), 25.3356491, 83.0076292),
            # ('Bengaluru', (12.9767936, 77.590082), 12.9767936, 77.590082),
            ('Delhi', (28.6273928, 77.1716954), 28.6273928, 77.1716954),
            ('Ladakh', (33.9456407, 77.6568576), 33.9456407, 77.6568576),
            # ('Mumbai', (18.9733536, 72.82810491917377), 18.9733536, 72.82810491917377), # Mumbai seems to move!
            ('Mysore', (12.3051828, 76.6553609), 12.3051828, 76.6553609),
        ]


    @unittest.skip("Not processing transcript ")
    def test_tidy_transcript(self):


        infile = Path(Resources.TEST_RESOURCES_DIR, "misc", "transcript1.txt")
        assert infile.exists(), f"infile should exists {infile}"
        with open(infile, "r") as f:
            lines = f.readlines()

        assert lines is not None
        assert len(lines) == pytest.approx(1800, abs=20)
        lines2 = []
        for ll in lines:
            ll2 = multiple_replace(subs, ll)
            if len(ll.strip()) > 0:
                lines2.append(ll2)
        transcript_dir = Path(Resources.TEMP_DIR, "misc", "transcript")
        transcript_dir.mkdir(exist_ok=True, parents=True)
        outfile = Path(transcript_dir, "transcript2.txt")
        with open(outfile, "w") as f:
            f.writelines(lines2)
        logger.debug(f"lines2 {lines2}")

    def test_word_counter(self):
        """
        illustratiue exercise
        reads a ChAPTER of IPCC, tokenize, lowercase ans count words in Counter

        """
        min_len = 5
        infile = Path(Resources.TEST_RESOURCES_DIR, "ipcc", "wg3", "Chapter03", "html_with_ids.html")
        assert infile.exists()
        container_xpath = ".//div[@id='executive-summary']"
        text_xpath = "div/p//text()"

        p_counter = self.count_words_crude(infile, container_xpath, min_len, text_xpath)
        logger.debug(f"counter {p_counter}")

    def count_words_crude(self, infile, container_xpath, min_len, text_xpath):
        """
        reads html file, extracts <p> objects , and counts lowercase text tokens
        :param container_xpath: top_level container
        ;param infile: html file
        :param min_len: of text chunks
        :param text_xpath;to find text chunks
        """
        htmlx = HtmlUtil.parse_html_file_to_xml(infile)
        exec_summ = htmlx.xpath(container_xpath)
        p_texts = exec_summ[0].xpath(text_xpath)
        p_counter = Counter()
        for p_text in p_texts:
            if len(p_text) >= min_len:
                p_text = p_text.lower().strip()
                splits = p_text.split()
                for split in splits:
                    p_counter[split] += 1
        return p_counter
